import logging
from contextlib import contextmanager
from functools import lru_cache
from typing import Generator

from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.pool import StaticPool

from ixmp4.conf.base import PlatformInfo
from ixmp4.conf.manager import ManagerConfig, ManagerPlatformInfo
from ixmp4.conf.user import User
from ixmp4.core.exceptions import ProgrammingError
from ixmp4.data.db import (
    BaseModel,
    DataPointRepository,
    IndexSetRepository,
    ModelRepository,
    RegionRepository,
    RunMetaEntryRepository,
    RunRepository,
    ScalarRepository,
    ScenarioRepository,
    TableRepository,
    TimeSeriesRepository,
    UnitRepository,
    VariableRepository,
)

from ..auth.context import AuthorizationContext
from .base import (
    Backend,
)
from .base import (
    IamcSubobject as BaseIamcSubobject,
)
from .base import (
    OptimizationSubobject as BaseOptimizationSubobject,
)

logger = logging.getLogger(__name__)


@lru_cache()
def cached_create_engine(dsn: str) -> Engine:
    logger.info(f"Creating database engine for {dsn}")
    return create_engine(dsn)


class IamcSubobject(BaseIamcSubobject):
    datapoints: DataPointRepository
    timeseries: TimeSeriesRepository
    variables: VariableRepository


class OptimizationSubobject(BaseOptimizationSubobject):
    indexsets: IndexSetRepository
    scalars: ScalarRepository
    tables: TableRepository


class SqlAlchemyBackend(Backend):
    iamc: IamcSubobject
    info: PlatformInfo
    meta: RunMetaEntryRepository
    models: ModelRepository
    optimization: OptimizationSubobject
    regions: RegionRepository
    runs: RunRepository
    scenarios: ScenarioRepository
    units: UnitRepository
    Session = sessionmaker(autocommit=False, autoflush=False, future=True)
    auth_context: AuthorizationContext | None = None

    def __init__(self, info: PlatformInfo) -> None:
        super().__init__(info)
        self.make_engine(info.dsn)
        self.make_repositories()

    def make_engine(self, dsn: str):
        if dsn.startswith("postgresql://"):
            logger.debug(
                "Replacing the platform dsn prefix to use the new `psycopg` driver."
            )
            dsn = dsn.replace("postgresql://", "postgresql+psycopg://")
        self.engine = cached_create_engine(dsn)
        self.session = self.Session(bind=self.engine)

    def make_repositories(self):
        self.iamc.datapoints = DataPointRepository(self)
        self.iamc.timeseries = TimeSeriesRepository(self)
        self.iamc.variables = VariableRepository(self)
        self.meta = RunMetaEntryRepository(self)
        self.models = ModelRepository(self)
        self.optimization.indexsets = IndexSetRepository(self)
        self.optimization.scalars = ScalarRepository(self)
        self.optimization.tables = TableRepository(self)
        self.regions = RegionRepository(self)
        self.runs = RunRepository(self)
        self.scenarios = ScenarioRepository(self)
        self.units = UnitRepository(self)

    def close(self):
        self.session.close()

    @contextmanager
    def auth(
        self,
        user: User,
        manager: ManagerConfig,
        info: ManagerPlatformInfo,
        overlap_ok: bool = False,
    ) -> Generator[AuthorizationContext, None, None]:
        if self.auth_context is not None and not overlap_ok:
            raise ProgrammingError("Overlapping auth context.")

        self.auth_context = AuthorizationContext(user, manager, info)
        yield self.auth_context
        self.auth_context = None

    def _create_all(self):
        BaseModel.metadata.create_all(bind=self.engine)

    def _drop_all(self):
        BaseModel.metadata.drop_all(bind=self.engine, checkfirst=True)

    def reset(self):
        self.session.commit()
        self._drop_all()
        self._create_all()


class SqliteTestBackend(SqlAlchemyBackend):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            PlatformInfo(name="sqlite-test", dsn="sqlite:///:memory:"),
            *args,
            **kwargs,
        )
        self.reset()

    def make_engine(self, dsn: str):
        self.engine = create_engine(
            dsn,
            connect_args={"check_same_thread": False},
            poolclass=StaticPool,
        )
        self.session = self.Session(bind=self.engine)


class PostgresTestBackend(SqlAlchemyBackend):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            PlatformInfo(
                name="postgres-test",
                dsn="postgresql://postgres:postgres@localhost/test",
            ),
            *args,
            **kwargs,
        )
        self.reset()



from datetime import datetime
from typing import ClassVar

import pandas as pd
from pydantic import Field

from ixmp4.data import abstract

from . import base
from .docs import Docs, DocsRepository


class RegionParent(base.BaseModel):
    id: int
    name: str
    hierarchy: str


class Region(base.BaseModel):
    NotFound: ClassVar = abstract.Region.NotFound
    NotUnique: ClassVar = abstract.Region.NotUnique
    DeletionPrevented: ClassVar = abstract.Region.DeletionPrevented

    id: int
    name: str
    hierarchy: str

    created_at: datetime | None = Field(None)
    created_by: str | None = Field(None)


class RegionDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/regions/"


class RegionRepository(
    base.Creator[Region],
    base.Deleter[Region],
    base.Retriever[Region],
    base.Enumerator[Region],
    abstract.RegionRepository,
):
    model_class = Region
    prefix = "regions/"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.docs = RegionDocsRepository(self.backend)

    def create(
        self,
        name: str,
        hierarchy: str,
    ) -> Region:
        return super().create(name=name, hierarchy=hierarchy)

    def delete(self, id: int) -> None:
        super().delete(id)

    def get(self, name: str) -> Region:
        return super().get(name=name)

    def enumerate(self, **kwargs) -> list[Region] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, **kwargs) -> list[Region]:
        return super()._list(json=kwargs)

    def tabulate(self, **kwargs) -> pd.DataFrame:
        return super()._tabulate(json=kwargs)



"""
This module holds a shared datastructure and interface for normalization
between the database and api data models and repositories.
"""
# flake8: noqa

from .base import (
    BaseMeta,
    BaseModel,
    BulkDeleter,
    BulkUpserter,
    Creator,
    Deleter,
    Enumerator,
    Lister,
    Retriever,
    Tabulator,
)
from .docs import Docs, DocsRepository
from .iamc import (  # AnnualDataPoint,; SubAnnualDataPoint,; CategoricalDataPoint,
    DataPoint,
    DataPointRepository,
    Measurand,
    MeasurandRepository,
    TimeSeries,
    TimeSeriesRepository,
    Variable,
    VariableRepository,
)
from .meta import MetaValue, RunMetaEntry, RunMetaEntryRepository, StrictMetaValue
from .model import Model, ModelRepository
from .optimization import (
    IndexSet,
    IndexSetRepository,
    Scalar,
    ScalarRepository,
    Table,
    TableRepository,
)
from .region import Region, RegionRepository
from .run import Run, RunRepository
from .scenario import Scenario, ScenarioRepository
from .unit import Unit, UnitRepository



from fastapi import APIRouter, Depends, Path, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination

router: APIRouter = APIRouter(
    prefix="/docs",
    tags=["docs"],
)


class DocsInput(BaseModel):
    dimension_id: int
    description: str


@router.get("/models/", response_model=EnumerationOutput[api.Docs])
def list_models(
    dimension_id: int | None = Query(None),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.models.docs.list(dimension_id=dimension_id),
        total=backend.models.docs.count(dimension_id=dimension_id),
        pagination=pagination,
    )


@router.post("/models/", response_model=api.Docs)
def set_models(
    docs: DocsInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.models.docs.set(**docs.model_dump())


@router.delete("/models/{dimension_id}/")
def delete_models(
    dimension_id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.models.docs.delete(dimension_id)


@router.get("/regions/", response_model=EnumerationOutput[api.Docs])
def list_regions(
    dimension_id: int | None = Query(None),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.regions.docs.list(dimension_id=dimension_id),
        total=backend.regions.docs.count(dimension_id=dimension_id),
        pagination=pagination,
    )


@router.post("/regions/", response_model=api.Docs)
def set_regions(
    docs: DocsInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.regions.docs.set(**docs.model_dump())


@router.delete("/regions/{dimension_id}/")
def delete_regions(
    dimension_id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.regions.docs.delete(dimension_id)


@router.get("/scenarios/", response_model=EnumerationOutput[api.Docs])
def list_scenarios(
    dimension_id: int | None = Query(None),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.scenarios.docs.list(dimension_id=dimension_id),
        total=backend.scenarios.docs.count(dimension_id=dimension_id),
        pagination=pagination,
    )


@router.post("/scenarios/", response_model=api.Docs)
def set_scenarios(
    docs: DocsInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.scenarios.docs.set(**docs.model_dump())


@router.delete("/scenarios/{dimension_id}/")
def delete_scenarios(
    dimension_id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.scenarios.docs.delete(dimension_id)


@router.get("/units/", response_model=EnumerationOutput[api.Docs])
def list_units(
    dimension_id: int | None = Query(None),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.units.docs.list(dimension_id=dimension_id),
        total=backend.units.docs.count(dimension_id=dimension_id),
        pagination=pagination,
    )


@router.post("/units/", response_model=api.Docs)
def set_units(
    docs: DocsInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.units.docs.set(**docs.model_dump())


@router.delete("/units/{dimension_id}/")
def delete_units(
    dimension_id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.units.docs.delete(dimension_id)


@router.get("/iamc/variables/", response_model=EnumerationOutput[api.Docs])
def list_variables(
    dimension_id: int | None = Query(None),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.iamc.variables.docs.list(dimension_id=dimension_id),
        total=backend.iamc.variables.docs.count(dimension_id=dimension_id),
        pagination=pagination,
    )


@router.post("/iamc/variables/", response_model=api.Docs)
def set_variables(
    docs: DocsInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.variables.docs.set(**docs.model_dump())


@router.delete("/iamc/variables/{dimension_id}/")
def delete_variables(
    dimension_id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.variables.docs.delete(dimension_id)


@router.get("/optimization/indexsets/", response_model=EnumerationOutput[api.Docs])
def list_indexsets(
    dimension_id: int | None = Query(None),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.optimization.indexsets.docs.list(dimension_id=dimension_id),
        total=backend.optimization.indexsets.docs.count(dimension_id=dimension_id),
        pagination=pagination,
    )


@router.post("/optimization/indexsets/", response_model=api.Docs)
def set_indexsets(
    docs: DocsInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.indexsets.docs.set(**docs.model_dump())


@router.delete("/optimization/indexsets/{dimension_id}/")
def delete_indexsets(
    dimension_id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.indexsets.docs.delete(dimension_id)


@router.get("/optimization/scalars/", response_model=list[api.Docs])
def list_scalars(
    dimension_id: int | None = Query(None),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.docs.list(dimension_id=dimension_id)


@router.post("/optimization/scalars/", response_model=api.Docs)
def set_scalars(
    docs: DocsInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.docs.set(**docs.model_dump())


@router.delete("/optimization/scalars/{dimension_id}/")
def delete_scalars(
    dimension_id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.scalars.docs.delete(dimension_id)


@router.get("/optimization/tables/", response_model=list[api.Docs])
def list_tables(
    dimension_id: int | None = Query(None),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.tables.docs.list(dimension_id=dimension_id)


@router.post("/optimization/tables/", response_model=api.Docs)
def set_tables(
    docs: DocsInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.tables.docs.set(**docs.model_dump())


@router.delete("/optimization/tables/{dimension_id}/")
def delete_tables(
    dimension_id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    return backend.optimization.tables.docs.delete(dimension_id)



from typing import ClassVar

from sqlalchemy import UniqueConstraint

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.db.optimization.indexset import IndexSet

from .. import base


class Column(base.BaseModel):
    NotFound: ClassVar = abstract.Column.NotFound
    NotUnique: ClassVar = abstract.Column.NotUnique
    DeletionPrevented: ClassVar = abstract.Column.DeletionPrevented

    name: types.String = db.Column(db.String(255), nullable=False, unique=False)
    # Currently not in use:
    dtype: types.String = db.Column(
        db.String(255), nullable=False, unique=False
    )  # pandas dtype

    table__id: types.Mapped[int] = db.Column(
        db.Integer, db.ForeignKey("optimization_table.id"), index=True
    )
    indexset: types.Mapped[IndexSet] = db.relationship(single_parent=True)
    constrained_to_indexset: types.Mapped[int] = db.Column(
        db.Integer, db.ForeignKey("optimization_indexset.id"), index=True
    )

    # Currently not in use:
    unique: types.Boolean = db.Column(db.Boolean, default=True)

    __table_args__ = (UniqueConstraint(name, table__id),)



from typing import Any

import numpy as np
import pandas as pd
import pandera as pa
from pandera.typing import DataFrame, Series

from ixmp4 import db
from ixmp4.core.decorators import check_types
from ixmp4.core.exceptions import InconsistentIamcType, ProgrammingError
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.model import Model
from ixmp4.data.db.region import Region
from ixmp4.data.db.run import Run, RunRepository
from ixmp4.data.db.scenario import Scenario
from ixmp4.data.db.unit import Unit

from .. import base
from ..measurand import Measurand
from ..timeseries import TimeSeries, TimeSeriesRepository
from ..variable import Variable
from . import get_datapoint_model
from .filter import DataPointFilter
from .model import DataPoint


class RemoveDataPointFrameSchema(pa.DataFrameModel):
    type: Series[pa.String] | None = pa.Field(isin=[t for t in DataPoint.Type])
    step_year: Series[pa.Int] | None = pa.Field(coerce=True, nullable=True)
    step_datetime: Series[pa.DateTime] | None = pa.Field(coerce=True, nullable=True)
    step_category: Series[pa.String] | None = pa.Field(nullable=True)

    time_series__id: Series[pa.Int] = pa.Field(coerce=True)

    @pa.dataframe_check
    def validate_type(cls, df: pd.DataFrame) -> bool:
        types = df.type.unique()

        # mixed types are currently not supported
        if len(types) != 1:
            raise InconsistentIamcType

        if types[0] == DataPoint.Type.ANNUAL:
            cols = dict(step_year="valid", step_category="empty", step_datetime="empty")
        elif types[0] == DataPoint.Type.CATEGORICAL:
            cols = dict(step_year="valid", step_category="valid", step_datetime="empty")
        elif types[0] == DataPoint.Type.DATETIME:
            cols = dict(step_year="empty", step_category="empty", step_datetime="valid")
        else:
            raise ProgrammingError

        for col, content in cols.items():
            if infer_content(df, col) is not content:
                raise InconsistentIamcType
        return True


class AddDataPointFrameSchema(RemoveDataPointFrameSchema):
    value: Series[pa.Float] = pa.Field(coerce=True)


class UpdateDataPointFrameSchema(AddDataPointFrameSchema):
    id: Series[pa.Int] = pa.Field(coerce=True)


def infer_content(df: pd.DataFrame, col: str) -> str:
    if col not in df.columns or all(pd.isna(df[col])):
        return "empty"
    elif not any(pd.isna(df[col])):
        return "valid"
    else:
        raise InconsistentIamcType


class DataPointRepository(
    base.Enumerator[DataPoint],
    base.BulkUpserter[DataPoint],
    base.BulkDeleter[DataPoint],
    abstract.DataPointRepository,
):
    model_class = DataPoint
    timeseries: TimeSeriesRepository
    runs: RunRepository

    def __init__(self, *args, **kwargs) -> None:
        backend, *_ = args
        # A different table was used for ORACLE databases (deprecated since ixmp4 0.3.0)
        self.model_class = get_datapoint_model(backend.session)

        self.timeseries = TimeSeriesRepository(*args, **kwargs)
        self.runs = RunRepository(*args, **kwargs)

        self.filter_class = DataPointFilter
        super().__init__(*args, **kwargs)

    def join_auth(self, exc: db.sql.Select) -> db.sql.Select:
        if not db.utils.is_joined(exc, TimeSeries):
            exc = exc.join(
                TimeSeries, onclause=self.model_class.time_series__id == TimeSeries.id
            )
        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return exc

    def select_joined_parameters(self, join_runs=False):
        bundle = []
        if join_runs:
            bundle.extend(
                [
                    Model.name.label("model"),
                    Scenario.name.label("scenario"),
                    Run.version.label("version"),
                ]
            )

        bundle.extend(
            [
                Region.name.label("region"),
                Unit.name.label("unit"),
                Variable.name.label("variable"),
                self.bundle,
            ]
        )

        _exc = (
            db.select(*bundle)
            .join(
                TimeSeries, onclause=self.model_class.time_series__id == TimeSeries.id
            )
            .join(Region, onclause=TimeSeries.region__id == Region.id)
            .join(Measurand, onclause=TimeSeries.measurand__id == Measurand.id)
            .join(Unit, onclause=Measurand.unit__id == Unit.id)
            .join(Variable, onclause=Measurand.variable__id == Variable.id)
        )

        if join_runs:
            _exc = (
                _exc.join(Run, onclause=TimeSeries.run__id == Run.id)
                .join(Model, onclause=Model.id == Run.model__id)
                .join(Scenario, onclause=Scenario.id == Run.scenario__id)
            )
        return _exc

    def select(
        self,
        *,
        join_parameters: bool | None = False,
        join_runs: bool = False,
        _filter: DataPointFilter | None = None,
        _exc: db.sql.Select | None = None,
        **kwargs: Any,
    ) -> db.sql.Select:
        if _exc is not None:
            exc = _exc
        elif join_parameters:
            exc = self.select_joined_parameters(join_runs)
        else:
            exc = db.select(self.bundle)

        return super().select(_exc=exc, _filter=_filter, **kwargs)

    @guard("view")
    def list(self, *args, **kwargs) -> list[DataPoint]:
        return super().list(*args, **kwargs)

    @guard("view")
    def tabulate(
        self, *args: Any, _raw: bool | None = False, **kwargs: Any
    ) -> pd.DataFrame:
        if _raw:
            return super().tabulate(*args, **kwargs)
        df = super().tabulate(*args, **kwargs)
        df["value"] = df["value"].astype(np.float64)
        df = df.sort_values(
            by=["time_series__id", "step_year", "step_category", "step_datetime"]
        )
        return df.dropna(axis="columns", how="all")

    def check_df_access(self, df: pd.DataFrame):
        if self.backend.auth_context is not None:
            ts_ids = set(df["time_series__id"].unique().tolist())
            self.timeseries.check_access(
                ts_ids,
                access_type="edit",
                run={"default_only": False, "is_default": None},
            )

    @check_types
    @guard("edit")
    def bulk_upsert(self, df: DataFrame[AddDataPointFrameSchema]) -> None:
        return super().bulk_upsert(df)

    @check_types
    @guard("edit")
    def bulk_insert(self, df: DataFrame[AddDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        return super().bulk_insert(df)

    @check_types
    @guard("edit")
    def bulk_update(self, df: DataFrame[UpdateDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        return super().bulk_update(df)

    @check_types
    @guard("edit")
    def bulk_delete(self, df: DataFrame[RemoveDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        res = super().bulk_delete(df)
        self.delete_orphans()
        return res

    def delete_orphans(self):
        exc = db.delete(TimeSeries).where(
            ~db.exists(
                db.select(self.model_class.id).where(
                    TimeSeries.id == self.model_class.time_series__id
                )
            )
        )
        self.session.execute(exc, execution_options={"synchronize_session": False})
        self.session.commit()



"""
This module is responsible for everything database related.
Interfacing, migrating and more is all done here.

It uses `sqlalchemy <https://www.sqlalchemy.org/>`__ and
`alembic <https://alembic.sqlalchemy.org/en/latest/>`__ for database
management.

Migrations
----------

There is a development database at ``run/db.sqlite`` which is used for
generating migrations, nothing else. It can be manipulated with alembic
directly using these commands:

.. code:: bash

   # run all migrations until the current state is reached
   alembic upgrade head

   # run one migration forward
   alembic upgrade +1

   # run one migration backward
   alembic downgrade -1

   # autogenerate new migration (please choose a descriptive change message)
   alembic revision -m "<message>" --autogenerate

You will have to run all migrations before being able to create new ones
in the development database. Be sure to run ``ruff`` on newly created
migrations before committing them!

"""
# flake8: noqa

from sqlalchemy import (
    ForeignKey,
    Index,
    Sequence,
    UniqueConstraint,
    delete,
    exists,
    func,
    insert,
    or_,
    select,
    sql,
    update,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.exc import MultipleResultsFound
from sqlalchemy.orm import (
    Relationship,
    Session,
    aliased,
    backref,
    mapped_column,
    relationship,
)
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.types import *

from . import utils

Column = mapped_column
JsonType = JSON()
JsonType = JsonType.with_variant(JSONB(), "postgresql")



from typing import Any

import numpy as np
import pandas as pd
from sqlalchemy import select
from sqlalchemy.orm import Bundle

from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.iamc.measurand import Measurand
from ixmp4.data.db.region import Region, RegionRepository
from ixmp4.data.db.run import RunRepository
from ixmp4.data.db.timeseries import (
    TimeSeriesRepository as BaseTimeSeriesRepository,
)
from ixmp4.data.db.unit import Unit, UnitRepository
from ixmp4.data.db.utils import map_existing

from ..measurand import MeasurandRepository
from ..variable import Variable
from .model import TimeSeries


class TimeSeriesRepository(
    BaseTimeSeriesRepository[TimeSeries], abstract.TimeSeriesRepository
):
    model_class = TimeSeries

    regions: RegionRepository
    measurands: MeasurandRepository
    units: UnitRepository

    def __init__(self, *args, **kwargs) -> None:
        from .filter import TimeSeriesFilter

        self.filter_class = TimeSeriesFilter

        self.runs = RunRepository(*args, **kwargs)
        self.regions = RegionRepository(*args, **kwargs)
        self.measurands = MeasurandRepository(*args, **kwargs)
        self.units = UnitRepository(*args, **kwargs)
        super().__init__(*args, **kwargs)

    @guard("view")
    def get(self, run_id: int, **kwargs: Any) -> TimeSeries:
        return super().get(run_id, **kwargs)

    def select_joined_parameters(self):
        return (
            select(
                self.bundle,
                Bundle(
                    "Region",
                    Region.name.label("region"),
                ),
                Bundle(
                    "Unit",
                    Unit.name.label("unit"),
                ),
                Bundle(
                    "Variable",
                    Variable.name.label("variable"),
                ),
            )
            .join(Region, onclause=TimeSeries.region__id == Region.id)
            .join(Measurand, onclause=TimeSeries.measurand__id == Measurand.id)
            .join(Unit, onclause=Measurand.unit__id == Unit.id)
            .join(Variable, onclause=Measurand.variable__id == Variable.id)
        )

    @guard("view")
    def list(self, **kwargs) -> list[TimeSeries]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs) -> pd.DataFrame:
        return super().tabulate(**kwargs)

    @guard("edit")
    def bulk_upsert(self, df: pd.DataFrame, create_related: bool = False) -> None:
        if self.backend.auth_context is not None:
            run_ids = set(df["run__id"].unique().tolist())
            self.runs.check_access(
                run_ids,
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        if create_related:
            df = self.map_relationships(df)
            df = df.drop_duplicates()
        super().bulk_upsert(df)

    def map_regions(self, df: pd.DataFrame):
        existing_regions = self.regions.tabulate(name__in=df["region"].unique())
        df, missing = map_existing(
            df,
            existing_df=existing_regions,
            join_on=("name", "region"),
            map=("id", "region__id"),
        )
        if len(missing) > 0:
            raise Region.NotFound(", ".join(missing))

        return df

    def map_measurands(self, df: pd.DataFrame) -> pd.DataFrame:
        df, missing = map_existing(
            df,
            existing_df=self.units.tabulate(),
            join_on=("name", "unit"),
            map=("id", "unit__id"),
        )
        if len(missing) > 0:
            raise Unit.NotFound(", ".join(missing))

        df["measurand__id"] = np.nan

        def map_measurand(df):
            variable_name, unit__id = df.name
            measurand = self.measurands.get_or_create(
                variable_name=variable_name, unit__id=int(unit__id)
            )
            df["measurand__id"] = measurand.id
            df["variable"] = variable_name
            df["unit__id"] = unit__id
            return df

        return pd.DataFrame(
            df.groupby(["variable", "unit__id"], group_keys=False).apply(
                map_measurand, include_groups=False
            )
        )

    def map_relationships(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self.map_regions(df)
        df = self.map_measurands(df)
        return df



from types import UnionType
from typing import Any, ClassVar, Optional, Union, get_args, get_origin

from pydantic import BaseModel, ConfigDict, Field, ValidationError, model_validator
from pydantic.fields import FieldInfo

from ixmp4 import db
from ixmp4.core.exceptions import BadFilterArguments, ProgrammingError


def exact(c, v):
    return c == v


def in_(c, v):
    return c.in_(v)


def like(c, v):
    return c.like(escape_wildcard(v), escape="\\")


def ilike(c, v):
    return c.ilike(escape_wildcard(v), escape="\\")


def notlike(c, v):
    return c.notlike(escape_wildcard(v), escape="\\")


def notilike(c, v):
    return c.notilike(escape_wildcard(v), escape="\\")


def gt(c, v):
    return c > v


def lt(c, v):
    return c < v


def gte(c, v):
    return c >= v


def lte(c, v):
    return c <= v


def escape_wildcard(v):
    return v.replace("%", "\\%").replace("*", "%")


class Integer(int):
    """An explicit proxy type for `int`."""

    pass


class Float(float):
    """An explicit proxy type for `float`."""

    pass


class Id(int):
    """A no-op type for a reduced set of `Integer` lookups."""

    pass


class String(str):
    """An explicit proxy type for `str`."""

    pass


Boolean = bool

argument_seperator = "__"
filter_func_prefix = "filter_"
lookup_map: dict[object, dict] = {
    Id: {
        "__root__": (int, exact),
        "in": (list[int], in_),
    },
    Float: {
        "__root__": (int, exact),
        "in": (list[int], in_),
        "gt": (int, gt),
        "lt": (int, lt),
        "gte": (int, gte),
        "lte": (int, lte),
    },
    Integer: {
        "__root__": (int, exact),
        "in": (list[int], in_),
        "gt": (int, gt),
        "lt": (int, lt),
        "gte": (int, gte),
        "lte": (int, lte),
    },
    Boolean: {
        "__root__": (bool, exact),
    },
    String: {
        "__root__": (str, exact),
        "in": (list[str], in_),
        "like": (str, like),
        "ilike": (str, ilike),
        "notlike": (str, notlike),
        "notilike": (str, notilike),
    },
}


def get_filter_func_name(n: str) -> str:
    return filter_func_prefix + n.strip()


PydanticMeta: type = type(BaseModel)


class FilterMeta(PydanticMeta):
    def __new__(cls, name: str, bases: tuple, namespace: dict, **kwargs):
        annots = namespace.get("__annotations__", {}).copy()
        for name, annot in annots.items():
            if get_origin(annot) == ClassVar:
                continue
            cls.process_field(namespace, name, annot)

        return super().__new__(cls, name, bases, namespace, **kwargs)

    @classmethod
    def build_lookups(cls, field_type: type) -> dict:
        global lookup_map
        if field_type not in lookup_map.keys():
            if get_origin(field_type) in [Union, UnionType]:
                unified_types = get_args(field_type)
                ut_lookup_map = {
                    type_: cls.build_lookups(type_) for type_ in unified_types
                }
                all_lookup_names = set()
                for tl in ut_lookup_map.values():
                    all_lookup_names |= set(tl.keys())

                lookups = {}
                for lookup_name in all_lookup_names:
                    tuples = [
                        ut_lookup_map[type_][lookup_name]
                        for type_ in unified_types
                        if ut_lookup_map[type_].get(lookup_name, None) is not None
                    ]
                    types, _ = zip(*tuples)

                    def lookup_func(c, v, tuples=tuples):
                        for t, lf in tuples:
                            if isinstance(v, t):
                                return lf(c, v)
                        raise ProgrammingError

                    lookups[lookup_name] = (
                        # dynamic union types can't
                        # be done according to type checkers
                        Union[tuple(types)],  # type:ignore
                        lookup_func,
                    )
                return lookups
            else:
                return {"__root__": (field_type, exact)}
        else:
            return lookup_map[field_type]

    @classmethod
    def process_field(cls, namespace: dict, field_name: str, field_type: type):
        lookups = cls.build_lookups(field_type)
        field: FieldInfo | None = namespace.get(field_name, Field(default=None))

        if field is None:
            return

        namespace.setdefault(field_name, field)
        override_lookups: list | None = None
        if isinstance(field.json_schema_extra, dict):
            jschema_lookups = field.json_schema_extra.get("lookups", None)
            if isinstance(jschema_lookups, list):
                override_lookups = jschema_lookups
            else:
                raise ProgrammingError(
                    "Field argument `lookups` must be `list` of `str`."
                )
        else:
            override_lookups = None
        if isinstance(override_lookups, list):
            lookups = {k: v for k, v in lookups.items() if k in override_lookups}
        elif override_lookups is None:
            pass
        else:
            lookups = {}
        base_field_alias = str(field.alias) if field.alias else field_name

        cls.expand_lookups(
            field_name,
            lookups,
            namespace,
            base_field_alias,
        )

    @classmethod
    def expand_lookups(
        cls,
        name: str,
        lookups: dict,
        namespace: dict,
        base_field_alias: str | None = None,
    ):
        global argument_seperator
        for lookup_alias, (type_, func) in lookups.items():
            if lookup_alias == "__root__":
                filter_name = name
            else:
                filter_name = name + argument_seperator + lookup_alias

            namespace["__annotations__"][filter_name] = Optional[type_]
            func_name = get_filter_func_name(filter_name)

            def filter_func(self, exc, f, v, func=func, session=None):
                return exc.where(func(f, v))

            namespace.setdefault(func_name, filter_func)

            if (
                base_field_alias is not None
                and base_field_alias != name
                and lookup_alias != "__root__"
            ):
                field = namespace.get(
                    filter_name,
                    Field(
                        None, alias=base_field_alias + argument_seperator + lookup_alias
                    ),
                )
            else:
                field = namespace.get(filter_name, Field(None))
            field.json_schema_extra = {"sqla_column": name}
            namespace[filter_name] = field


class BaseFilter(BaseModel, metaclass=FilterMeta):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        populate_by_name=True,
    )
    sqla_model: ClassVar[type | None] = None

    @model_validator(mode="before")
    @classmethod
    def expand_simple_filters(cls, v):
        return expand_simple_filter(v)

    def __init__(self, **data: Any) -> None:
        try:
            super().__init__(**data)

        except ValidationError as e:
            raise BadFilterArguments(model=e.title, errors=e.errors())

    def join(self, exc: db.sql.Select, session=None) -> db.sql.Select:
        return exc

    def apply(self, exc: db.sql.Select, model, session) -> db.sql.Select:
        dict_model = dict(self)
        for name, field_info in self.model_fields.items():
            value = dict_model.get(name, field_info.get_default())
            if isinstance(value, BaseFilter):
                submodel = getattr(value, "sqla_model", None)
                model_getter = getattr(value, "get_sqla_model", None)
                if submodel is None and callable(model_getter):
                    submodel = model_getter(session)

                exc = value.join(exc, session=session)
                exc = value.apply(exc, submodel, session)
            elif value is not None:
                func_name = get_filter_func_name(name)
                filter_func = getattr(self, func_name, None)
                if filter_func is None:
                    raise ProgrammingError

                sqla_column: str | None = None
                if isinstance(field_info.json_schema_extra, dict):
                    jschema_col = field_info.json_schema_extra.get("sqla_column")
                    if not isinstance(jschema_col, str):
                        raise ProgrammingError(
                            "Field argument `sqla_column` must be of type `str`."
                        )
                    sqla_column = jschema_col
                else:
                    sqla_column = None
                if sqla_column is None:
                    column = None
                else:
                    column = getattr(model, sqla_column, None)
                exc = filter_func(exc, column, value, session=session)
        return exc.distinct()


def expand_simple_filter(value):
    if isinstance(value, str):
        if "*" in value:
            return dict(name__like=value)
        else:
            return dict(name=value)
    elif isinstance(value, list):
        if any(["*" in v for v in value]):
            raise NotImplementedError("Filter by list with wildcard is not implemented")
        return dict(name__in=value)

    return value



from typing import Protocol

import pandas as pd

from ixmp4.data import types

from . import base
from .docs import DocsRepository


class Unit(base.BaseModel, Protocol):
    """Unit data model."""

    name: types.String
    "Unique name of the unit."
    created_at: types.DateTime
    "Creation date/time. TODO"
    created_by: types.String
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<Unit {self.id} name={self.name}>"


class UnitRepository(
    base.Creator,
    base.Deleter,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(self, name: str) -> Unit:
        """Creates a unit.

        Parameters
        ----------
        name : str
            The name of the model.

        Raises
        ------
        :class:`ixmp4.core.exceptions.UnitNotUnique`:
            If the unit with `name` is not unique.

        Returns
        -------
        :class:`ixmp4.data.abstract.Unit`:
            The created unit.
        """
        ...

    def get(self, name: str) -> Unit:
        """Retrieves a unit.

        Parameters
        ----------
        name : str
            The unique name of the unit.

        Raises
        ------
        :class:`ixmp4.data.abstract.Unit.NotFound`:
            If the unit with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Unit`:
            The retrieved unit.
        """
        ...

    def get_by_id(self, id: int) -> Unit:
        """Retrieves a Unit by it's id.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.Unit.NotFound`.
            If the Unit with `id` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Unit`:
            The retrieved Unit.
        """
        ...

    def get_or_create(self, name: str) -> Unit:
        try:
            return self.get(name)
        except Unit.NotFound:
            return self.create(name)

    def delete(self, id: int):
        """Deletes a unit.

        Parameters
        ----------
        id : int
            The unique integer id of the unit.

        Raises
        ------
        :class:`ixmp4.data.abstract.Unit.NotFound`:
            If the unit with `id` does not exist.
        :class:`ixmp4.data.abstract.Unit.DeletionPrevented`:
            If the unit with `id` is used in the database, preventing it's deletion.
        """
        ...

    def list(
        self,
        *,
        name: str | None = None,
        **kwargs,
    ) -> list[Unit]:
        r"""Lists units by specified criteria.

        Parameters
        ----------
        name : str
            The name of a unit. If supplied only one result will be returned.
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.unit.filters.UnitFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.base.iamc.Unit`]:
            List of units.
        """
        ...

    def tabulate(
        self,
        *,
        name: str | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        r"""Tabulate units by specified criteria.

        Parameters
        ----------
        name : str
            The name of a unit. If supplied only one result will be returned.
        \*\*kwargs: any
            More filter parameters as specified in
            `ixmp4.data.db.unit.filters.UnitFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:pass
                - id
                - name
        """
        ...



from ixmp4.data.abstract import Run

from ..base import BaseFacade
from .indexset import IndexSetRepository
from .scalar import ScalarRepository
from .table import TableRepository


class OptimizationData(BaseFacade):
    """An optimization data instance, which provides access to optimization data such as
    IndexSet, Table, Variable, etc."""

    indexsets: IndexSetRepository
    scalars: ScalarRepository
    tables: TableRepository

    def __init__(self, *args, run: Run, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.indexsets = IndexSetRepository(_backend=self.backend, _run=run)
        self.scalars = ScalarRepository(_backend=self.backend, _run=run)
        self.tables = TableRepository(_backend=self.backend, _run=run)



from fastapi import APIRouter, Body, Depends, Path, Query

from ixmp4.core.exceptions import BadRequest
from ixmp4.data import abstract, api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.meta.filter import RunMetaEntryFilter

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination

router: APIRouter = APIRouter(
    prefix="/meta",
    tags=["meta"],
)


class RunMetaEntryInput(BaseModel):
    run__id: int
    key: str
    value: abstract.StrictMetaValue


@router.patch("/", response_model=EnumerationOutput[api.RunMetaEntry])
def query(
    filter: RunMetaEntryFilter = Body(None),
    join_run_index: bool | None = Query(False),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    if join_run_index and not table:
        raise BadRequest("`join_run_index` can only be used with `table=true`.")

    return EnumerationOutput(
        results=backend.meta.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
            join_run_index=join_run_index,
        ),
        total=backend.meta.count(_filter=filter),
        pagination=pagination,
    )


@router.post("/", response_model=api.RunMetaEntry)
def create(
    runmeta: RunMetaEntryInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.meta.create(**runmeta.model_dump())


@router.delete("/{id}/")
def delete(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
):
    backend.meta.delete(id)


@router.post("/bulk/")
def bulk_upsert(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.meta.bulk_upsert(df.to_pandas())


@router.patch("/bulk/")
def bulk_delete(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.meta.bulk_delete(df.to_pandas())



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.iamc.variable.filter import VariableFilter

from .. import deps
from ..base import BaseModel, EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/variables",
    tags=["iamc", "variables"],
)


class VariableInput(BaseModel):
    name: str


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.Variable])
def query(
    filter: VariableFilter = Body(VariableFilter()),
    table: bool = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
):
    return EnumerationOutput(
        results=backend.iamc.variables.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.iamc.variables.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.post("/", response_model=api.Variable)
def create(
    variable: VariableInput,
    backend: Backend = Depends(deps.get_backend),
):
    return backend.iamc.variables.create(**variable.model_dump())
